/*                                                                -*- mode: c; c-basic-offset: 8;  -*-
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */



#include "Common.h"
#include "Definitions.h"
#include "gfal_is_ifce_internal.h"



char *default_search_base = NULL;

char *plugin_is_set() {
    char *temp = getenv("GLITE_SD_PLUGIN");
    if (temp) {
        if (strlen(temp) > 0) {
            return strdup(temp);
        }
    }
    return NULL;
}

char *base_is_set(void) {
    char *temp = getenv("LCG_GFAL_BDII_DEFAULT_SEARCH_BASE");
    if (temp) {
        if (strlen(temp) > 0) {
            return strdup(temp);
        }
    }
    return NULL;
}

int gfal_is_interface_test_get_max_uri_length() {
    return MAX_URI_LENGTH;
}

char *strtok_uri(char *uri) {
    const char *proto = "ldap://";
    const int proto_len = strlen(proto);

    if (uri == NULL) {
        return NULL;
    }

    /* In case of empty string, return a dynamically allocated empty string */
    if (uri[0] == 0) {
        char *ret = malloc(1);
        ret[0] = 0;
        return ret;
    }

    {
        int uri_len = strlen(uri);
        int uri_with_first_proto_len = uri_len + proto_len;
        char *uri_with_first_proto = malloc(uri_with_first_proto_len + 1);
        char *uri_with_protos = NULL;
        char *replacement = ",ldap://";
        GFAL_IS_IFCE_ASSERT(uri_with_first_proto);
        uri_with_first_proto[0] = 0;
        /* Insert "ldap://" to the beginning of the string */
        strcat(uri_with_first_proto, proto);
        strcat(uri_with_first_proto, uri);
        /* Replace each "," to ",ldap://" - it effectively inserts the protocol 
           in front of the host names */
        uri_with_protos = gfal_is_ifce_replace_substr(uri_with_first_proto, ",", replacement);
        free(uri_with_first_proto);
        return uri_with_protos;
    }
}


void close_connection(LDAP *ld) {
    int saverrno = errno;
    if (ld) {
        ldap_unbind_ext_s(ld, NULL, NULL);
        ld = NULL;
    }
    errno = saverrno;
}

/* Return an LDAP connection to the BDII server */
LDAP *get_connection(SDException *exception, char *errbuf, int errbufsz) {

    char *lcg_infosys = NULL;
    int ret = 0;
    LDAP *ldcache = NULL;
    struct timeval timeout;

    default_search_base = sbasedn;

    lcg_infosys = getenv("LCG_GFAL_INFOSYS");
    if (!lcg_infosys) {
        errno = EINVAL;
        if (exception)
            SD_setException(exception, SDStatus_FAILURE, "LCG_GFAL_INFOSYS is not set", errbuf, errbufsz);
        return NULL;
    }

    /* ldap_initialize() can handle the "url://host:port" syntax just fine */
    /*It requires the full uri to be specified*/
    char url[MAX_URI_LENGTH] = {0};
    char *strtok_uri_var = strtok_uri(lcg_infosys);
    strncpy(url, strtok_uri_var, strlen(strtok_uri_var));
    free(strtok_uri_var);
    ret = ldap_initialize(&ldcache, url);
    if (ret != LDAP_SUCCESS) {
        char buf[256];
        snprintf(buf, sizeof (buf), "ldap_initialize() failed: %s", ldap_err2string(ret));
        if (exception)
            SD_setException(exception, SDStatus_FAILURE, buf, errbuf, errbufsz);
        return NULL;
    }

    if (get_timeout() > 0) {
        timeout.tv_sec = get_timeout();
        timeout.tv_usec = 0;
        ldap_set_option(ldcache, LDAP_OPT_NETWORK_TIMEOUT, &timeout);
    }

    struct berval cred;
    cred.bv_val = NULL;
    cred.bv_len = 0;

    ret = ldap_sasl_bind_s(ldcache, NULL, LDAP_SASL_SIMPLE, &cred, NULL, NULL, NULL);
    if (ret != LDAP_SUCCESS) {
        char buf[256];
        snprintf(buf, sizeof (buf), "ldap_sasl_bind_s() failed: %s", ldap_err2string(ret));
        if (exception)
            SD_setException(exception, SDStatus_FAILURE, buf, errbuf, errbufsz);
        close_connection(ldcache);
        return NULL;
    }
    return ldcache;
}

void report_query_error(const char *fmt, LDAP *ld, LDAPMessage *msg, int err,
        SDException *exception) {
    char *error = NULL;
    int ret = 0;
    int errcodep = 0;
    char *error_msg = NULL;

    if (NULL != msg && ld != NULL) {
        ret = ldap_parse_result(ld, msg, &errcodep, NULL, &error_msg, NULL, NULL, TRUE);
    } else {
        ret = err;
    }
    /* Verify that the result was parsed correctly. */
    if (ret != LDAP_SUCCESS) {
        error = g_strdup_printf(fmt, ldap_err2string(ret));

        SD_setException(exception, SDStatus_FAILURE, error, NULL, 0);

        g_free(error);
        if (error_msg)
            ldap_memfree(error_msg);
        return;
    }
    /* Check the results of the operation. */
    if (errcodep != LDAP_SUCCESS) {
        error = g_strdup_printf(fmt, ldap_err2string(errcodep));

        SD_setException(exception, SDStatus_FAILURE, error, NULL, 0);

        g_free(error);
        if (error_msg)
            ldap_memfree(error_msg);

    }
}

char *gfal_is_interafce_test_strtok_uri(char *uri) {
    return strtok_uri(uri);
}

void SD_setException(SDException *exception, int status, const char *reason, char *errbuf, int errbufsz) {
    if (!exception) {
        return;
    }

    exception->status = status;
    if (status == SDStatus_SUCCESS) {
        exception->reason = NULL;
    } else {
        if (reason != NULL) {
            exception->reason = (char *) malloc(strlen(reason) + 1);
            if (exception->reason != NULL) {
                strcpy(exception->reason, reason);
            } else {
                exception->reason = SD_NOMEM_MESSAGE;
            }
        } else {
            exception->reason = SD_EMPTY_MESSAGE;
        }
    }

    if (status != SDStatus_SUCCESS && errbufsz != 0) {
        strncpy(errbuf, reason, errbufsz - 1);
        errbuf[errbufsz - 1] = '\0'; /*nul terminated*/
        SD_freeException(exception);
    }
}

/* Get the value of a single-valued attribute */
char *get_single_value(LDAP *ld, LDAPMessage *msg, const char *attr) {

    struct berval **values = NULL;
    char *res = NULL;

    values = ldap_get_values_len(ld, msg, attr);
    if (!values) {
        return NULL;
    }

    if (values && values[0] && values[0]->bv_val) {
        res = g_strdup(values[0]->bv_val);
        ber_bvecfree(values);
    }
    return res;
}

/**
 * Utility function, connects and searched the ldap server
 * Param	filter/query, attr to be searched, pointer to an ldap server
 * Return	LDAPMessage pointer
 */

int connect_search_ldap(char *base, char *query, char *attr[], LDAP *ld,
        SDException *exception, LDAPMessage **reply, char *errbuf, int errbufsz) {

    struct timeval timeout;
    int ret = 0;

    timeout.tv_sec = get_timeout();
    timeout.tv_usec = 0;

    ret = ldap_search_ext_s(ld, base, LDAP_SCOPE_SUBTREE, query, attr, 0, NULL, NULL, &timeout, 0, reply);
    if (ret != LDAP_SUCCESS) {
        if (ret == LDAP_TIMELIMIT_EXCEEDED || ret == LDAP_TIMEOUT) {
            SD_setException(exception, SDStatus_FAILURE, "[ldap_search_st][]: Connection Timeout", errbuf, errbufsz);
        } else {
            SD_setException(exception, SDStatus_FAILURE, ldap_err2string(ret), errbuf, errbufsz);
        }
    }
    return ret;
}

int get_timeout() {
    char *bdii_env = getenv("LCG_GFAL_BDII_TIMEOUT");
    if (bdii_env != NULL) {
        return atoi(bdii_env);
    }
    return QUERY_TIMEOUT;
}

const char *get_vo(const char *acrb) {
    char * ret = 0;
    char *start = strstr(acrb, "VO:");

    if (NULL != start) {
        ret = start + 3;
        return ret;
    }

    return acrb;
}

char *get_host(const char *endpoint) {

    char * ret;
    char *start = strstr(endpoint, "://");

    if (NULL != start) {
        ret = start + 3;
    } else {
        ret = (char*) endpoint;
    }

    char *end = strchr(ret, ':');
    if (NULL != end) {
        *end = '\0';
    } else {
        end = strchr(ret, '/');
        if (NULL != end)
            *end = '\0';
    }

    return ret;
}

char *dupstr(const char *from) {
    char *to;

    to = (char *) malloc(strlen(from) + 1);
    if (to != NULL) strcpy(to, from);
    return to;
}

SDVOList *checkVOEnv(void) {
    SDVOList *volist;
    const char *vo;

    vo = getenv(SDEnv_VO);
    if (!vo)
        return NULL;

    if (strlen(vo) == 0)
        return NULL;

    volist = (SDVOList *) calloc(1, sizeof (SDVOList));
    if (!volist)
        return NULL;

    volist->names = (char **) malloc(sizeof (char *));
    if (!volist->names) {
        free(volist);
        return NULL;
    }

    volist->numNames = 1;
    volist->names[0] = dupstr(vo);
    if (!volist->names[0]) {
        free(volist->names);
        free(volist);
        return NULL;
    }
    return volist;
}

void SD_freeVOList(SDVOList *vos) {
    int i;

    if (!vos)
        return;
    for (i = 0; i < vos->numNames; i++)
        g_free(vos->names[i]);
    g_free(vos->names);
    g_free(vos);
}

void parse_url(const char *url,
        char **scheme, char **host, int *port, char **path) {
    char *p = NULL, *q = NULL;

    *scheme = *host = *path = 0;
    *port = -1;

    /* Does it start with a scheme? */

    for (p = (char *) url; *p; p++)
        if (*p == ':' || *p == '/')
            break;

    if (p > url && *p == ':') {
        *scheme = malloc(p - url + 1);
        strncpy(*scheme, url, p - url);
        (*scheme)[p - url] = '\0';
        url = p + 1;
    }

    /* Does it have a net_loc? */

    if (url[0] == '/' && url[1] == '/') {
        url += 2;

        for (p = (char *) url; *p; p++)
            if (*p == '/')
                break;

        /* Does it have a port number? */

        for (q = p - 1; q >= url; q--)
            if (!isdigit(*q))
                break;

        if (q < p - 1 && *q == ':')
            *port = atoi(q + 1);
        else
            q = p;

        *host = malloc(q - url + 1);
        strncpy(*host, url, q - url);
        (*host)[q - url] = '\0';
        url = p;
    }

    /* The rest is the path */

    if (*url)
        *path = strdup(url);
    else
        *path = strdup("/");

    for (p = *path; *p; p++)
        if (*p == '\\') {
            *p = '/';
        }
}

char *mystrcasestr(const char *haystack, const char *needle) {
#ifdef HAVE_STRCASESTR
    return strcasestr(haystack, needle);
#else
    size_t haystacks;
    size_t needles;
    const char *p;

    if (NULL == needle || '\0' == *needle) {
        return (char *) haystack;
    }
    if (NULL == haystack || '\0' == *haystack) {
        return NULL;
    }
    haystacks = strlen(haystack);
    needles = strlen(needle);
    if (haystacks < needles) {
        return NULL;
    }

    for (p = haystack; p <= &haystack[haystacks - needles]; p++) {
        if (0 == strncasecmp(p, needle, needles)) {
            return (char *) p;
        }
    }
    return NULL;
#endif
}
